#!/usr/bin/python
"""
This module provides some basic helper/formatting utilities,
specifically targeted at decoding ReadHistoryData data.

"""
import io
from binascii import hexlify

import lib
from records import *
from datetime import date

_remote_ids = [
  bytearray([ 0x01, 0xe2, 0x40 ]),
  bytearray([ 0x03, 0x42, 0x2a ]),
  bytearray([ 0x0c, 0x89, 0x92 ]),
]

def decode_remote_id(msg):
  """
  practice decoding some remote ids:

  | 0x27
  | 0x01 0xe2 0x40
  | 0x03 0x42 0x2a
  | 0x28 0x0c 0x89
  | 0x92 0x00 0x00 0x00

  >>> decode_remote_id(_remote_ids[0])
  '123456'

  >>> decode_remote_id(_remote_ids[1])
  '213546'

  >>> decode_remote_id(_remote_ids[2])
  '821650'



  """
  high   = msg[ 0 ] * 256 * 256
  middle = msg[ 1 ] * 256
  low    = msg[ 2 ]
  return str(high + middle + low)

class AlarmPump(KnownRecord):
  opcode = 0x06
  head_length = 4
#class ResultTotals(KnownRecord):
class ResultDailyTotal(InvalidRecord):
  """On 722 this seems like two records."""
  opcode = 0x07
  #head_length = 5
  head_length = 5
  date_length = 2
  #body_length = 37 + 4
  #body_length = 2
  def __init__(self, head, larger=False):
    super(type(self), self).__init__(head, larger)
    if self.larger:
      self.body_length = 3
  def parse_time(self):
    date = parse_midnight(self.date)
    self.datetime = date
    if not hasattr(date, 'isoformat'):
      self.datetime = None
    return date

  def decode (self):
    self.parse_time( )
    mid = unmask_m_midnight(self.date)[0:3]
    return (dict(valid_date=date(*mid).isoformat()))

  def date_str(self):
    result = 'unknown'
    if self.datetime is not None:
      result = self.datetime.isoformat( )
    else:
      if len(self.date) >=2:
        result = "{}".format(unmask_m_midnight(self.date))
    return result

class ChangeBasalProfile_old_profile (KnownRecord):
  opcode = 0x08
  # old/bad
  # body_length = 46
  # XXX: on LeDanaScott's, 522, this seems right
  body_length = 145
  # head_length = 3 # XXX: # for 554!?
  def __init__(self, head, larger=False):
    super(type(self), self).__init__(head, larger)
    if self.larger:
      self.body_length = 145
  def decode (self):
    self.parse_time( )
    rates = [ ]
    i = 0
    for x in range(47):
      start = x * 3
      end = start + 3
      (offset, rate, q) = self.body[start:end]
      if [offset, rate, q] == [ 0x00, 0x00, 0x00]:
        break
      try:
        rates.append(describe_rate(offset, rate, q))
      except TypeError, e:
        remainder = [ offset, rate, q ]
        rates.append(remainder)
    return rates

def describe_rate (offset, rate, q):
  return (dict(offset=(30*1000*60)*offset, rate=rate/40.0))


class ChangeBasalProfile_new_profile (KnownRecord):
  opcode = 0x09
  body_length = 145
  # body_length = 144 # XXX: # for 554!?
  # head_length = 3 # XXX: # for 554!?
  def decode (self):
    self.parse_time( )
    rates = [ ]
    i = 0
    for x in range(47):
      start = x * 3
      end = start + 3
      (offset, rate, q) = self.body[start:end]
      if [offset, rate, q] == [ 0x00, 0x00, 0x00]:
        break
      rates.append(describe_rate(offset, rate, q))
    return rates
class ClearAlarm(KnownRecord):
  opcode = 0x0C
class SelectBasalProfile(KnownRecord):
  opcode = 0x14
class ChangeTime(KnownRecord):
  opcode = 0x17
class NewTimeSet(KnownRecord):
  opcode = 0x18
class LowBattery(KnownRecord):
  opcode = 0x19
class Battery(KnownRecord):
  opcode = 0x1a
class PumpSuspend(KnownRecord):
  opcode = 0x1e
class PumpResume(KnownRecord):
  opcode = 0x1f

class Rewind(KnownRecord):
  opcode = 0x21

class EnableDisableRemote(KnownRecord):
  opcode = 0x26
  # body_length = 14
  # head_length = 3 # XXX: for 554
  body_length = 14
class ChangeRemoteID(KnownRecord):
  opcode = 0x27

class TempBasalDuration(KnownRecord):
  opcode = 0x16
  _test_1 = bytearray([ ])
  def decode(self):
    self.parse_time( )
    basal = { 'duration (min)': self.head[1] * 30, }
    return basal
class ChangeMazaheri2e (KnownRecord):
  opcode = 0x2e
  body_length = 100

# class BolusWizard512 (BolusWizard):
class BolusWizard512 (KnownRecord):
  opcode = 0x2f
  body_length = 12

class UnabsorbedInsulin512 (UnabsorbedInsulinBolus):
  opcode = 0x30

class TempBasal (KnownRecord):
  opcode = 0x33
  body_length = 1
  _test_1 = bytearray([ ])
  def decode(self):
    self.parse_time( )
    temp = { 0: 'absolute', 1: 'percent' }[(self.body[0] >> 3)]
    status = dict(temp=temp)
    if temp is 'absolute':
      rate = self.head[1] / 40.0
      status.update(rate=rate)
    if temp is 'percent':
      rate = int(self.head[1])
      status.update(rate=rate)
    return status

class LowReservoir(KnownRecord):
  """
  >>> rec = LowReservoir( LowReservoir._test_1[:2] )
  >>> decoded = rec.parse(LowReservoir._test_1)
  >>> print str(rec)
  LowReservoir 2012-12-07T11:02:43 head[2], body[0] op[0x34]

  >>> print pformat(decoded)
  {'amount': 20.0}
  """
  opcode = 0x34
  _test_1 = bytearray([ 0x34, 0xc8,
                        0xeb, 0x02, 0x0b, 0x07, 0x0c, ])
  def decode(self):
    self.parse_time( )
    reservoir = {'amount' : int(self.head[1]) / 10.0 }
    return reservoir


class ChangeAlarmNotifyMode (KnownRecord):
  opcode = 0x63
  body_length = 0
class ChangeTimeDisplay(KnownRecord):
  opcode = 0x64

class ChangeBolusWizardSetup (KnownRecord):
  opcode = 0x4f
  body_length = 40

_confirmed = [ Bolus, Prime, AlarmPump, ResultDailyTotal,
               ChangeBasalProfile_old_profile,
               ChangeBasalProfile_new_profile,
               ClearAlarm, SelectBasalProfile, TempBasalDuration, ChangeTime,
               NewTimeSet, LowBattery, Battery, PumpSuspend,
               PumpResume, CalBGForPH, Rewind, EnableDisableRemote,
               ChangeRemoteID, TempBasal, LowReservoir, BolusWizard,
               UnabsorbedInsulinBolus, ChangeAlarmNotifyMode, ChangeTimeDisplay,
               ChangeBolusWizardSetup, ]

# _confirmed.append(DanaScott0x09)
_confirmed.append(ChangeMazaheri2e)
_confirmed.append(BolusWizard512)
_confirmed.append(UnabsorbedInsulin512)


class JournalEntryMealMarker(KnownRecord):
  """Capture Event > Meal marker"""
  opcode = 0x40
  body_length = 2

  def decode(self):
    super(JournalEntryMealMarker, self).decode()

    return dict(carb_input=int(lib.BangInt([self.head[1], self.body[0]])))

_confirmed.append(JournalEntryMealMarker)

class JournalEntryExerciseMarker(KnownRecord):
  """Capture Event > Exercise marker"""
  opcode = 0x41
  body_length = 1
_confirmed.append(JournalEntryExerciseMarker)

class JournalEntryOtherMarker(KnownRecord):
  """Capture Event > Other"""
  opcode = 0x43
  body_length = 0
_confirmed.append(JournalEntryOtherMarker)

class Ian69(KnownRecord):
  opcode = 0x69
  body_length = 2
_confirmed.append(Ian69)

class Ian50(KnownRecord):
  opcode = 0x50
  body_length = 34

  # XXX: tghoward testing on 723 at length 30
  body_length = 30
  def __init__ (self, head, model, **kwds):
    super(Ian50, self).__init__(head, model, **kwds)
    self.body_length = model.Ian50Body
_confirmed.append(Ian50)

class Ian54(KnownRecord):
  opcode = 0x54
  body_length = 34 + 23
  body_length = 57
_confirmed.append(Ian54)

class AlarmSensor (KnownRecord):
  """Glucose sensor alarms.
    The second byte of the head represents the alarm type.
    The third byte contains an alarm-specific value.

    For example, a "Low Glucose" alarm type is:
    [
      0x0b,  # 11: Opcode
      0x66,  # 102: Low glucose subtype
      0x50   # 80: Glucose level (For a pump configured to mg/dL)
    ]
    """
  opcode = 0x0B
  head_length = 3

  alarm_types = {
    101: 'High Glucose',
    102: 'Low Glucose',
    104: 'Meter BG Now',
    105: 'Cal Reminder',
    106: 'Calibration Error',
    107: 'Sensor End',
    112: 'Weak Signal',
    113: 'Lost Sensor',
    115: 'Low Glucose Predicted'
  }

  def decode(self):
    super(AlarmSensor, self).decode()

    alarm_type = self.head[1]

    decoded_dict = {
      'alarm_type': alarm_type,
      'alarm_description': self.alarm_types.get(alarm_type, 'Unknown sensor alarm ({})'.format(alarm_type))
    }

    if alarm_type in (101, 102,):
      year_bits = extra_year_bits(self.date[4])
      decoded_dict['amount'] = int(lib.BangInt([year_bits[0], self.head[2]]))

    return decoded_dict
_confirmed.append(AlarmSensor)

class BGReceived (KnownRecord):
  opcode = 0x3F
  body_length = 3
  def decode (self):
    self.parse_time( )
    bg = (self.head[1] << 3) + (self.date[2] >> 5)
    return dict(link=str(self.body).encode('hex'), amount=bg)
_confirmed.append(BGReceived)

class IanA8(KnownRecord):
  opcode = 0xA8
  head_length = 10
_confirmed.append(IanA8)

class BasalProfileStart(KnownRecord):
  opcode = 0x7b
  body_length = 3
  def __init__(self, head, larger=False):
    super(type(self), self).__init__(head, larger)
    if self.larger:
      # body_length = 1
      pass
      # self.body_length = 48
  def decode (self):
    self.parse_time( )
    if (len(self.body) % 3 == 0):
      rate = describe_rate(*self.body)
      rate['profile_index'] = self.head[1]
      return rate
    else:
      return dict(raw=hexlify(self.body))
_confirmed.append(BasalProfileStart)

# 123, 143
class OldBolusWizardChange (KnownRecord):
  opcode = 0x5a
  body_length = 117
  def __init__(self, head, larger=False):
    super(type(self), self).__init__(head, larger)
    if self.larger:
      self.body_length = 117 + 17 + 3
      pass
  def decode (self):
    self.parse_time( )
    half = (self.body_length - 1) / 2
    stale = self.body[0:half]
    changed = self.body[half:-1]
    tail = self.body[-1]
    return dict(stale=decode_wizard_settings(stale, model=self.model)
    # , _changed=changed
    , changed=decode_wizard_settings(changed, model=self.model)
    , tail=tail
    )

_confirmed.append(OldBolusWizardChange)
def decode_wizard_settings (data, num=8, model=None):
  head = data[0:2]
  tail = data[len(head):]
  carb_reader = model.read_carb_ratios.msg
  cr_size = carb_reader.item_size
  carb_ratios = tail[0:num*cr_size]
  tail = tail[num*cr_size:]
  insulin_sensitivies = tail[0:(num*2)]
  tail = tail[num*2:]
  isMg = head[0] & 0b00000100
  isMmol = head[0] & 0b00001000
  bg_units = 1
  if isMmol and not isMg:
    bg_units = 2
  bg_targets = tail[0:(num*3)+2]
  if model and model.larger:
    bg_targets = bg_targets[2:]
  return dict(head=str(head).encode('hex')
  # , carb_ratios=decode_carb_ratios(carb_ratios)
  , carb_ratios=carb_reader.decode_ratios(carb_ratios)
  # , _carb_ratios=str(carb_ratios).encode('hex')
  # , cr_len=len(carb_ratios)
  , insulin_sensitivies=decode_insulin_sensitivies(insulin_sensitivies)
  # , _insulin_sensitivies=str(insulin_sensitivies).encode('hex')
  # , is_len=len(insulin_sensitivies)
  # , bg_len=len(bg_targets)
  , bg_targets=decode_bg_targets(bg_targets, bg_units)
  # , _o_len=len(data)
  # , _bg_targets=str(bg_targets).encode('hex')
  , _head = "{0:#010b} {1:#010b}".format(*head)
  )

def decode_carb_ratios (data):
  ratios = [ ]
  for x in range(8):
    start = x * 3
    end = start + 3
    (offset, q, r) = data[start:end]
    ratio = r/10.0
    if q:
      ratio = lib.BangInt([q, r]) / 1000.0
    ratios.append(dict(i=x, offset=offset*30, q=q, _offset=offset,
                       ratio=ratio, r=r))
  return ratios

def decode_insulin_sensitivies (data):
  sensitivities = [ ]
  for x in range(8):
    start = x * 2
    end = start + 2
    (offset, sensitivity) = data[start:end]
    sensitivities.append(dict(i=x, offset=offset*30, _offset=offset,
                       sensitivity=sensitivity))
  return sensitivities

def decode_bg_targets (data, bg_units):
  # data = data[2:]
  targets = [ ]
  for x in range(8):
    start = x * 3
    end = start + 3
    # (low, high, offset) = data[start:end]
    (offset, low, high) = data[start:end]
    if bg_units is 2:
      low = low / 10.0
      high = high / 10.0
    targets.append(dict( #i=x,
                       offset=offset*30, _offset=offset,
                       # _raw=str(data[start:end]).encode('hex'),
                       low=low, high=high))
  return targets

class BigBolusWizardChange (KnownRecord):
  opcode = 0x5a
  body_length = 143

class SetAutoOff (KnownRecord):
  opcode = 0x1b
_confirmed.append(SetAutoOff)

class ChangeAudioBolus (KnownRecord):
  opcode = 0x5f
  def decode (self):
    self.parse_time( )
_confirmed.append(ChangeAudioBolus)

class ChangeCaptureEventEnable (KnownRecord):
  opcode = 0x83
  # body_length = 1
_confirmed.append(ChangeCaptureEventEnable)

class hack53 (KnownRecord):
  opcode = 0x53
  body_length = 1
_confirmed.append(hack53)

class hack52 (KnownRecord):
  opcode = 0x52
  # body_length = 1
_confirmed.append(hack52)

class hack51 (KnownRecord):
  opcode = 0x51
  # body_length = 1
_confirmed.append(hack51)

class hack55 (KnownRecord):
  opcode = 0x55
  # body_length = 1 + 47
  # body_length = 2 + 46
  def __init__(self, head, larger=False):
    super(type(self), self).__init__(head, larger)
    # self.larger = larger
    self.body_length = (self.head[1] - 1) * 3
_confirmed.append(hack55)


class hack56 (KnownRecord):
  opcode = 0x56
  body_length = 5
_confirmed.append(hack56)

class ChangeWatchdogMarriageProfile(KnownRecord):
  opcode = 0x81
  body_length = 5
_confirmed.append(ChangeWatchdogMarriageProfile)

class DeleteOtherDeviceID (KnownRecord):
  opcode = 0x82
  body_length = 5
_confirmed.append(DeleteOtherDeviceID)

class ChangeOtherDeviceID (KnownRecord):
  opcode = 0x7d
  body_length = 30
_confirmed.append(ChangeOtherDeviceID)

class SetBolusWizardEnabled (KnownRecord):
  opcode = 0x2d
  def decode (self):
    self.parse_time( )
    return dict(enabled=self.head[1] is 1)
_confirmed.append(SetBolusWizardEnabled)


class SettingSomething57 (KnownRecord):
  opcode = 0x57
  # body_length = 1
_confirmed.append(SettingSomething57)

class ChangeMaxBasal (KnownRecord):
  opcode = 0x2c
  def decode (self):
    self.parse_time( )
    return dict(maxBasal=self.head[1] / 40.0)
_confirmed.append(ChangeMaxBasal)

class questionable22 (KnownRecord):
  opcode = 0x22
_confirmed.append(questionable22)

class questionable23 (KnownRecord):
  opcode = 0x23
_confirmed.append(questionable23)

class questionable24 (KnownRecord):
  opcode = 0x24
_confirmed.append(questionable24)

class ChangeBGReminderEnable (KnownRecord):
  opcode = 0x60
  def decode (self):
    self.parse_time( )
    enabled = self.head[1] is 1
    return dict(enabled=enabled)
_confirmed.append(ChangeBGReminderEnable)

class questionable61 (KnownRecord):
  opcode = 0x61
_confirmed.append(questionable61)

class ChangeTempBasalType (KnownRecord):
  opcode = 0x62
  def decode (self):
    self.parse_time( )
    temp = { 0: 'absolute', 1: 'percent' }[self.head[1]]
    return dict(temp=temp)
  # body_length = 1
_confirmed.append(ChangeTempBasalType)

class questionable65 (KnownRecord):
  opcode = 0x65
_confirmed.append(questionable65)

class questionable66 (KnownRecord):
  opcode = 0x66
_confirmed.append(questionable66)

class questionable6f (KnownRecord):
  opcode = 0x6f
_confirmed.append(questionable6f)

class questionable5e (KnownRecord):
  opcode = 0x5e
_confirmed.append(questionable5e)

class ChangeParadigmLinkID (KnownRecord):
  opcode = 0x3c
  body_length = 14
  def decode (self):
    self.parse_time( )
    data = self.body[1:]
    links = [ ]
    links.append(str(data[0:3]).encode('hex'))
    links.append(str(data[3:6]).encode('hex'))
    links.append(str(data[7:10]).encode('hex'))
    return dict(links=links)
_confirmed.append(ChangeParadigmLinkID)


class ConnectDevicesOtherDevicesEnabled (KnownRecord):
  opcode = 0x7c
  def decode(self):
    super(ConnectDevicesOtherDevicesEnabled, self).decode()
    return dict(enabled=self.head[1] == 1)

_confirmed.append(ConnectDevicesOtherDevicesEnabled)

class Model522ResultTotals(KnownRecord):
  opcode = 0x6d
  head_length = 1
  date_length = 2
  body_length = 40
  def parse_time(self):
    date = parse_midnight(self.date)
    self.datetime = date
    if not hasattr(date, 'isoformat'):
      self.datetime = None
    return date

  def date_str(self):
    result = 'unknown'
    if self.datetime is not None:
      result = self.datetime.isoformat( )
    else:
      if len(self.date) >=2:
        result = "{}".format(unmask_m_midnight(self.date))
    return result

# class Model522ResultTotals(KnownRecord):
class old6c(Model522ResultTotals):
  opcode = 0x6c
  #head_length = 45
  #xxx non 515
  # body_length = 38
  # body_length = 34
  # XXX: 515 only?
  # body_length = 31
  def __init__ (self, head, model, **kwds):
    super(old6c, self).__init__(head, model, **kwds)
    self.body_length = model.old6cBody + 3
_confirmed.append(old6c)

class questionable3b (KnownRecord):
  opcode = 0x3b
_confirmed.append(questionable3b)


from dateutil.relativedelta import relativedelta
def parse_midnight (data):
    mid = unmask_m_midnight(data)
    oneday = relativedelta(days=1)
    try:
      date = datetime(*mid) + oneday
      return date
    except ValueError, e:
      print "ERROR", e, lib.hexdump(data)
      pass
    return mid

def unmask_m_midnight(data):
  """
  Extract date values from a series of bytes.
  Always returns tuple given a bytearray of at least 3 bytes.

  Returns 6-tuple of scalar values year, month, day, hours, minutes,
  seconds.

  """
  data = data[:]
  seconds = 0
  minutes = 0
  hours   = 0

  day     = parse_day(data[0])

  high = data[0] >> 4
  low  = data[0] & 0x1F

  year_high = data[1] >> 4
  # month = int(high) #+ year_high
  # month   = parse_months( data[0], data[1] )
  mhigh = (data[0] & 0xE0) >> 4
  mlow  = (data[1] & 0x80) >> 7
  month =  int(mhigh + mlow)
  day = int(low)

  year = parse_years(data[1])
  return (year, month, day, hours, minutes, seconds)

_confirmed.append(Model522ResultTotals)

class Sara6E(Model522ResultTotals):
  """Seems specific to 722?"""
  opcode = 0x6e
  #head_length = 52 - 5
  # body_length = 1
  body_length = 48
  #body_length = 0
  def __init__(self, head, larger=False):
    super(type(self), self).__init__(head, larger)
    if self.larger:
      self.body_length = 48
  def decode (self):
    self.parse_time( )
    mid = unmask_m_midnight(self.date)[0:3]
    try:
      return (dict(valid_date=date(*mid).isoformat()))
    except ValueError, e:
      return (dict(error_date=mid, error=str(e)))

_confirmed.append(Sara6E)

_known = { }

_variant = { }

for x in _confirmed:
  _known[x.opcode] = x

del x

def suggest(head, larger=False, model=None):
  """
  Look in the known table of commands to find a suitable record type
  for this opcode.
  """
  klass = _known.get(head[0], Base)
  record = klass(head, model)
  return record

def parse_record(fd, head=bytearray( ), larger=False, model=None):
  """
  Given a file-like object, and the head of a record, parse the rest
  of the record.
  Look up the type of record, read in just enough data to parse it,
  return the result.
  """
  # head    = bytearray(fd.read(2))
  date    = bytearray( )
  body    = bytearray( )
  record  = suggest(head, larger, model=model)
  remaining = record.head_length - len(head)
  if remaining > 0:
    head.extend(bytearray(fd.read(remaining)))
  if record.date_length > 0:
    date.extend(bytearray(fd.read(record.date_length)))
  if record.body_length > 0:
    body.extend(bytearray(fd.read(record.body_length)))
  record.parse( head + date + body )
  # print str(record)
  # print record.pformat(prefix=str(record) )
  return record


def describe( ):
  keys = _known.keys( )
  out  = [ ]
  for k in keys:
    out.append(_known[k].describe( ))
  return out

class PagedData (object):
  """
    PagedData - context for parsing a page of cgm data.
  """

  def __init__ (self, raw, model):
    self.model = model
    data, crc = raw[0:1022], raw[1022:]
    computed = lib.CRC16CCITT.compute(bytearray(data))
    if lib.BangInt(crc) != computed:
      assert lib.BangInt(crc) == computed, "CRC does not match page data"

    self.raw = raw
    self.clean(data)

  def clean (self, data):
    data.reverse( )
    self.data = self.eat_nulls(data)
    self.stream = io.BufferedReader(io.BytesIO(self.data))

  def eat_nulls (self, data):
    i = 0
    while data[i] == 0x00:
      i = i+1
    return data[i:]

class HistoryPage (PagedData):
  def clean (self, data):
    # data.reverse( )
    # self.data = self.eat_nulls(data)
    #self.data.reverse( )
    self.data = data[:]
    # XXX: under some circumstances, zero is the correct value and
    # eat_nulls actually eats valid data.  This ugly hack restores two
    # nulls back ot the end.
    """
    self.data.append(0x00)
    self.data.append(0x00)
    self.data.append(0x00)
    self.data.append(0x00)
    self.data.append(0x00)
    """
    self.stream = io.BufferedReader(io.BytesIO(self.data))
  def decode (self, larger=False):
    records = [ ]
    skipped = [ ]
    for B in iter(lambda: bytearray(self.stream.read(2)), bytearray("")):
      if B == bytearray( [ 0x00, 0x00 ] ):
        if skipped:
          if len(records) > 0:
            last = records[-1]
            last.update(appended=last.get('appended', [ ]) + skipped)
          # records.extend(skipped)
          skipped = [ ]
        break
      record = parse_record(self.stream, B, larger=larger, model=self.model)
      data = record.decode( )
      if record.datetime:
        rec = dict(timestamp=record.datetime.isoformat( ),
                   _type=str(record.__class__.__name__),
                   _body=lib.hexlify(record.body),
                   _head=lib.hexlify(record.head),
                   _date=lib.hexlify(record.date),
                   _description=str(record))
        if data is not None:
          rec.update(data)
          if skipped:
            rec.update(appended=skipped)
            skipped = [ ]
        records.append(rec)
      else:
        rec = dict(_type=str(record.__class__.__name__),
                   _body=lib.hexlify(record.body),
                   _head=lib.hexlify(record.head),
                   _date=lib.hexlify(record.date),
                   _description=str(record))
        data = record.decode( )
        if data is not None:
          rec.update(data=data)
        skipped.append(rec)
    records.reverse( )
    return records

if __name__ == '__main__':
  import doctest
  doctest.testmod( )

#####
# EOF




#
# TODO: move all constants to config module.
#

import serial
import logging
import lib
import fuser
io  = logging.getLogger( )
log = io.getChild(__name__)

class AlreadyInUseException (Exception):
  pass

class Link( object ):
  __timeout__ = .500
  port = None
  def __init__( self, port, timeout=None ):
    if timeout is not None:
      self.__timeout__ = timeout
    if fuser.in_use(port):
      raise AlreadyInUseException("{port} already in use".format(port=port))
    self.open( port, dsrdtr=True, rtscts=True )


  def open( self, newPort=False, **kwds ):
    if newPort:
      self.port = newPort
    if 'timeout' not in kwds:
      kwds['timeout'] = self.__timeout__
    kwds['rtscts'] = True
    kwds['dsrdtr'] = True

    self.serial = serial.Serial( self.port, **kwds )

    if self.serial.isOpen( ):
      log.info( '{agent} opened serial port: {serial}'\
         .format( serial = repr( self.serial ),
                  agent  =self.__class__.__name__ ) )

  def close( self ):
    io.info( 'closing serial port' )
    return self.serial.close( )

  def write( self, string ):
    r = self.serial.write( string )
    io.info( 'usb.write.len: %s\n%s' % ( len( string ),
                                         lib.hexdump( bytearray( string ) ) ) )
    return r

  def read( self, c ):
    r = self.serial.read( c )
    io.info( 'usb.read.len: %s'   % ( len( r ) ) )
    io.info( 'usb.read.raw:\n%s' % ( lib.hexdump( bytearray( r ) ) ) )
    return r

  def readline( self ):
    r = self.serial.readline( )
    io.info( 'usb.read.len: %s\n%s' % ( len( r ),
                                        lib.hexdump( bytearray( r ) ) ) )
    return r

  def readlines( self ):
    r = self.serial.readlines( )
    io.info( 'usb.read.len: %s\n%s' % ( len( r ),
                                        lib.hexdump( bytearray( ''.join( r ) ) ) ) )
    return r

if __name__ == '__main__':
  import doctest
  doctest.testmost( )


#####
# EOF



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK

from setuptools import setup, find_packages
import platform

import decocare

def readme():
    with open("README.markdown") as f:
        return f.read()

setup(name='decocare',
    version='0.0.28', # http://semver.org/
    description='Audit, inspect, and command MM insulin pumps.',
    long_description=readme(),
    author="Ben West",
    author_email="bewest+insulaudit@gmail.com",
    url="https://github.com/bewest/decoding-carelink",
    #namespace_packages = ['insulaudit'],
    packages=find_packages( ),
    install_requires = [
      'pyserial', 'python-dateutil', 'argcomplete'
    ],
    scripts = [
      'bin/mm-press-key.py',
      'bin/mm-send-comm.py',
      'bin/mm-set-suspend.py',
      'bin/mm-temp-basals.py',
      'bin/mm-decode-history-page.py',
      'bin/mm-latest.py',
      'bin/mm-bolus.py',
      'bin/mm-set-rtc.py',
      'bin/mm-pretty-csv',
    ],
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries'
    ],
    include_package_data=True,
    package_data = {
      'decocare': ['etc/*', '*.rules' ],
      # 'decocare.etc': ['*.rules' ],
    },
    data_files = [
      ('/etc/udev/rules.d', ['decocare/etc/80-medtronic-carelink.rules'] )
    ],
    zip_safe=False
)

#####
# EOF



from base import KnownRecord, VariableHead
from decocare import lib
# from .. import lib
from times import extra_year_bits
from pprint import pformat

class Bolus(KnownRecord):
  """
  >>> rec = Bolus(Bolus._test_1[:4])
  >>> decoded = rec.parse(Bolus._test_1)
  >>> print str(rec)
  Bolus 2012-12-18T15:05:28 head[4], body[0] op[0x01]

  >>> print pformat(decoded)
  {'amount': 5.6, 'duration': 0, 'programmed': 5.6, 'type': 'normal'}

  """
  _test_1 = bytearray([ 0x01, 0x38, 0x38, 0x00,
                        0xdc, 0x05, 0x4f, 0x12, 0x0c, ])
  opcode = 0x01
  head_length = 4
  def __init__(self, head, larger=False):
    super(Bolus, self).__init__(head, larger)
    # self.larger = larger
    if self.larger:
      self.head_length = 8
  def decode(self):
    self.parse_time( )
    dose = {
             'amount': self.head[2]/10.0,
             'programmed': self.head[1]/10.0,
             'duration': self.head[3] * 30,
             'type': self.head[3] > 0 and 'square' or 'normal'
           }
    if self.larger:
      duration = self.head[7] * 30
      dose = { 'amount': lib.BangInt(self.head[3:5])/40.0,
               'programmed':  lib.BangInt(self.head[1:3])/40.0,
               'unabsorbed': lib.BangInt(self.head[5:7])/40.0,
               'duration': duration,
               'type': duration > 0 and 'square' or 'normal',
             }
    return dose

class BolusWizard(KnownRecord):
  """
  Decode/parse bolus wizard records.

  >>> from decocare import models
  >>> rec = BolusWizard(BolusWizard._test_1[:2], model=models.PumpModel('522', None))
  >>> decoded = rec.parse(BolusWizard._test_1)
  >>> print str(rec)
  BolusWizard 2013-01-20T13:07:45 head[2], body[13] op[0x5b]

  >>> print pformat(decoded)
  {'_byte[5]': 0,
   '_byte[7]': 0,
   'bg': 108,
   'bg_target_high': 125,
   'bg_target_low': 106,
   'bolus_estimate': 1.1,
   'carb_input': 15,
   'carb_ratio': 13,
   'correction_estimate': 0.0,
   'food_estimate': 1.1,
   'sensitivity': 45,
   'unabsorbed_insulin_count': '??',
   'unabsorbed_insulin_total': 4.8,
   'unknown_byte[10]': 0,
   'unknown_byte[8]': 0}

  """
  # missing unabsorbed_insulin_count = 4
  _test_1 = bytearray([ 0x5b, 0x6c,
                        0x2d, 0x47, 0x0d, 0x14, 0x0d,
                        0x0f, 0x50, 0x0d, 0x2d, 0x6a, 0x00, 0x0b, 0x00,
                        0x00, 0x30, 0x00, 0x0b, 0x7d, ])

  _test_2 = bytearray([ 0x5b, 0x8b,
                        0xdc, 0x05, 0x0f, 0x12, 0x0c,
                        0x45, 0x50, 0x0d, 0x2d, 0x6a, 0x03, 0x35, 0x00,
                        0x00, 0x00, 0x00, 0x38, 0x7d, ])
  opcode = 0x5b
  body_length = 13
  def __init__(self, head, model=None):
    super(BolusWizard, self).__init__(head, model)
    # self.larger = larger
    self.MMOL_DEFAULT = model.MMOL_DEFAULT
    if self.larger:
      self.body_length = 15
  def decode(self):
    self.parse_time( )
    bg = lib.BangInt([ self.body[1] & 0x0f, self.head[1] ])
    carb_input = int(self.body[0])
    # XXX: I have no idea if this is correct; it seems to produce correct results.
    correction = ( twos_comp( self.body[7], 8 )
                 + twos_comp( self.body[5] & 0x0f, 8 ) ) / 10.0
    wizard = { 'bg': bg, 'carb_input': carb_input,
               'carb_ratio': int(self.body[2]),
               'sensitivity': int(self.body[3]),
               'bg_target_low': int(self.body[4]),
               'bg_target_high': int(self.body[12]),
               'bolus_estimate': int(self.body[11])/10.0,
               'food_estimate': int(self.body[6])/10.0,
               'unabsorbed_insulin_total': int(self.body[9])/10.0,
               'unabsorbed_insulin_count': '??',
               'correction_estimate': correction,
               '_byte[5]': self.body[5],
               '_byte[7]': int(self.body[7]), #
               'unknown_byte[8]': self.body[8],
               'unknown_byte[10]': self.body[10],
               # '??': '??',
               # 'unabsorbed_insulin_total': int(self.body[9])/10.0,
               # 'food_estimate': int(self.body[0]),
             }

    if self.larger:
      # correction = ( twos_comp( self.body[6], (self.body[9] & 0x38) << 5 ) ) / 40.0
      bg = ((self.body[1] & 0x03) << 8) + self.head[1]
      carb_input = ((self.body[1] & 0x0c) << 6) + self.body[0]
      carb_ratio = (((self.body[2] & 0x07) << 8) + self.body[3]) / 10.0
      # xxx: not sure about this
      # https://github.com/ps2/minimed_rf/blob/master/lib/minimed_rf/log_entries/bolus_wizard.rb#L102
      sensitivity = int(self.body[4])
      wizard = { 'bg': bg, 'carb_input': carb_input,
                 'carb_ratio': carb_ratio,
                 'sensitivity': sensitivity,
                 'bg_target_low': int(self.body[5]),
                 'bg_target_high': int(self.body[14]),
                 # 'bolus_estimate': int(self.body[13])/40.0,

                 'correction_estimate': (((self.body[9] & 0x38) << 5) + self.body[6]) / 40.0,
                 # 'correction_maybe_estimate': correction,

                 'food_estimate': insulin_decode(self.body[7], self.body[8]),
                 'unabsorbed_insulin_total': insulin_decode(self.body[10], self.body[11]),
                 'bolus_estimate': insulin_decode(self.body[12], self.body[13]),
                 # 'unknown_bytes': map(int, list(self.body)),
               }

    if self.MMOL_DEFAULT:
      for key in [ 'bg', 'bg_target_high', 'bg_target_low', 'sensitivity' ]:
        wizard[key] = wizard[key] / 10.0
    return wizard

def insulin_decode (a, b, strokes=40.0):
  return ((a << 8) + b) / strokes
def twos_comp(val, bits):
    # http://stackoverflow.com/a/9147327
    """compute the 2's compliment of int value val"""
    if( (val&(1<<(bits-1))) != 0 ):
        val = val - (1<<bits)
    return val

"""
moved to models
def decode_unabsorbed(amount, age, curve,strokes=40.0):
  unabsorbed = { 'amount': amount/strokes,
                 'age': age,
                 'curve': curve, }
  return unabsorbed
"""

class UnabsorbedInsulinBolus(VariableHead):
  """
  This data is not made available at the time of therapy in the pump
  UI, but could easily change my dosing decision.

  >>> from decocare import models
  >>> model = models.PumpModel('522', None)
  >>> rec = UnabsorbedInsulinBolus( UnabsorbedInsulinBolus._test_1[:2], model)
  >>> print str(rec)
  UnabsorbedInsulinBolus unknown head[2], body[0] op[0x5c]

  >>> print pformat(rec.parse( UnabsorbedInsulinBolus._test_1 ))
  [{'age': 78, 'amount': 1.25}, {'age': 88, 'amount': 0.95}]

  >>> rec = UnabsorbedInsulinBolus( UnabsorbedInsulinBolus._test_2[:2], model )
  >>> print str(rec)
  UnabsorbedInsulinBolus unknown head[2], body[0] op[0x5c]

  >>> print pformat(rec.parse( UnabsorbedInsulinBolus._test_2 ))
  [{'age': 60, 'amount': 2.6}, {'age': 160, 'amount': 2.5}]



  [{'age': 60, 'amount': 2.6, 'curve': 4},
   {'age': 160, 'amount': 2.5, 'curve': 4}]

  """
  _test_1 = bytearray([ 0x5c, 0x08,
                        0x32, 0x4e, 0x04, 0x26, 0x58, 0x04, ])

  _test_2 = bytearray([ 0x5c, 0x08,
                        0x68, 0x3c, 0x04,
                        0x64, 0xa0, 0x04, ])
  opcode = 0x5c
  date_length = 0
  def decode(self):
    raw = self.head[2:]
    return self.model.decode_unabsorbed(raw)

class CalBGForPH(KnownRecord):
  """
    >>> rec = CalBGForPH( CalBGForPH._test_1[:2] )
    >>> rec.parse( CalBGForPH._test_1 )
    {'amount': 139}
    >>> print str(rec)
    CalBGForPH 2012-12-18T15:04:46 head[2], body[0] op[0x0a]

  """
  _test_1 = bytearray([ 0x0a, 0x8b,
                        0xee, 0x04, 0x2f, 0x12, 0x0c, ])

  _test_2 = bytearray([ 0x0a, 0xa7,
                        0x22, 0x53, 0x30, 0x0e, 0x0d, ])

  _test_3 = bytearray([ 0x0a, 0xb0,
                        0x00, 0x6f, 0x2f, 0x0e, 0x0d, ])

  _test_4 = bytearray([ 0x0a, 0x42,
                        0x0c, 0x6c, 0x31, 0x0e, 0x8d, ])

  _test_5 = bytearray([ 0x0a, 0x60,
                        0x04, 0x59, 0x2b, 0x0e, 0x8d, ])
  _test_6 = bytearray([ 0x0a, 0x5b,
                        0x16, 0x52, 0x2a, 0x0e, 0x8d, ])
  opcode = 0x0a
  def decode(self):
    self.parse_time( )
    year_bits = extra_year_bits(self.date[4])

    return { 'amount': int(lib.BangInt([ year_bits[0], self.head[1] ])) }
    pass

if __name__ == '__main__':
  import doctest
  doctest.testmod( )



from datetime import datetime
#from .. import lib
from decocare import lib
from times import *

class Base(object):
  """
    >>> str( Base( bytearray([ 0x00, 0x00 ]) ) )
    'Base unknown head[2], body[0] op[0x00]'

  Each record in the history seems to have a two byte head, possibly
  some arguments, then a 5 byte description of the datetime, then
  maybe a body.  The most reliable way to identify records so far,
  seems to be through the 2 byte head.

  """
  head_length = 2
  body_length = 0
  date_length = 5
  def __init__(self, head=bytearray( ), model=None, larger=False):
    self.larger = getattr(model, 'larger', larger)
    self.model = model
    self.bolus = bytearray( )
    self.opcode = head[0]
    self.head = head
    self.date = bytearray( )
    self.datetime = None
    self.body = bytearray( )

  @classmethod
  def describe(klass):
    opstring = "0x%02x" % (getattr(klass, 'opcode', 0))
    name = klass.__name__
    out = [ klass.head_length, klass.date_length, klass.body_length ]
    return ",".join([ opstring, name ] + map(str, out))

  def __str__(self):
    name = self.__class__.__name__
    lengths = 'head[{}], body[{}]'.format(len(self.head), len(self.body))
    # opcodes = ' '.join(['%#04x' % x for x in self.head[:1]])
    opcodes = 'op[%#04x]' % self.opcode
    return ' '.join([name, self.date_str( ), lengths, opcodes ])

  def date_str(self):
    result = 'unknown'
    if self.datetime is not None:
      result = self.datetime.isoformat( )
    else:
      if len(self.date) == 5:
        result = "{}".format(unmask_date(self.date))
    return result

  def min_length(self):
    return self.head_length + self.date_length
  def parse(self, bolus):
    if len(bolus) < self.min_length( ):
      return
    head_length = self.head_length
    date_length = self.date_length

    self.bolus  = bolus
    self.head   = bolus[:head_length]
    body_offset = head_length + date_length
    self.date   = bolus[head_length:body_offset]
    self.body   = bolus[body_offset:]
    return self.decode( )

  def decode(self):
    pass

  def pformat(self, prefix=''):
    head = '\n'.join([ "    op hex (%s)" % len(self.head), lib.hexdump(self.head, indent=4),
                       "    decimal", lib.int_dump(self.head, indent=11) ])
    date = '\n'.join([ "    datetime (%s)" % self.date_str( ),
                       lib.hexdump(self.date, indent=4) ])

    body = "    body (%s)" % len(self.body)
    if len(self.body) > 0:
      body = '\n'.join([ body,
                         "    hex", lib.hexdump(self.body, indent=4),
                         "    decimal", lib.int_dump(self.body, indent=11) ])
    extra = [ ]
    """
    hour_bits = self.date[1:] and extra_hour_bits(self.date[1]) or [ ]
    year_bits = self.date[4:] and extra_year_bits(self.date[4]) or [ ]
    day_bits  = self.date[3:] and extra_hour_bits(self.date[3]) or [ ]
    if 1 in hour_bits:
      extra.append("HOUR BITS: {}".format(str(hour_bits)))
    if 1 in day_bits:
      extra.append("DAY BITS: {}".format(str(day_bits)))
    if 1 in year_bits:
      extra.append("YEAR BITS: {}".format(str(year_bits)))
    """
    decoded = None
    if len(self.head + self.date + self.body) >= self.min_length( ):
      decoded = self.decode( )
    decode_msg = ''
    if decoded is not None:
      decode_msg = '\n'.join([ '###### DECODED',
                               '```python',
                               '{}'.format(lib.pformat(self.decode( ))),
                               '```', ])
    if extra:
      extra = '    ' + ' '.join(extra)
    else:
      extra = ''
    return '\n'.join([ prefix, decode_msg, head, date, body, extra ])


class VariableHead(Base):
  def __init__(self, head, model=None):
    Base.__init__(self, head, model)
    self.head_length = head[1]

class KnownRecord(Base):
  opcode = 0x00
  decodes_date = True

  def parse_time(self):
    if len(self.date) == 5:
      self.datetime = parse_date(self.date)

  def decode(self):
    self.parse_time( )



class InvalidRecord(KnownRecord):
  pass
  # 0x0c: 22,
  # 0x6d: 46,
  # 0x6d: 46 - 5,

class Prime(KnownRecord):
  opcode = 0x03
  head_length = 5
  def decode(self):
    self.parse_time( )
    amount = self.head[4] / 10.0
    fixed  = self.head[2] / 10.0
    t = {0:'manual'}.get(fixed, 'fixed')
    prime = { 'type': t,
              'amount': amount,
              'fixed': fixed, }
    return prime

if __name__ == '__main__':
  import doctest
  doctest.testmod( )




from decocare import commands, history, cgm
from decocare import lib
import types

class Task (object):
  def __init__ (self, msg, handler=None, **kwargs):
    self.msg = msg
    if handler:
      self.func = handler
  def __get__ (self, obj, objtype=None):
    if obj is None:
      return self
    else:
      return types.MethodType(self, obj)
  def validate (self):
    data = self.response.getData( )
    self.response.check_output(data)
  @staticmethod
  def func (self, response):
    return response.getData( )
  def __call__ (self, inst, **kwds):
    # print "__calll__", inst, self.func
    # self.func( )
    self.response = inst.session.query(self.msg, **kwds)
    self.validate( )
    return types.MethodType(self.func, inst)(self.response)
    # return self.response.getData( )

  @classmethod
  def handler (klass, msg, **kwargs):
    def closure (func):
      return Task(msg, handler=func, **kwargs)
    return closure

class Cursor (object):
  # Info
  # Page
  def __init__ (self, inst, **kwds):
    self.inst = inst
    self.kwds = kwds
  def get_page_info (self):
    self.info = self.inst.session.query(self.Info)
  def download_page (self, num):
    page = self.inst.session.query(self.Page, page=num)
    for record in self.find_records(page):
      yield record
  def range (self, info):
    raise NotImplemented( )
  def find_records (self, response):
    raise NotImplemented( )
  def iter (self):
    self.get_page_info( )
    for n in self.range(self.info.getData( )):
      yield self.download_page(n)

class PageIterator (Task):

  def __init__ (self, Cursor=None, handler=None):
    self.Cursor = Cursor
    if handler:
      self.func = handler

  def __call__ (self, inst, **kwds):
    self.pager = self.Cursor(inst, **kwds)
    for page in self.pager.iter( ):
      for record in page:
        yield record

  @classmethod
  def handler (klass, **kwargs):
    def closure (func):
      return klass(func, **kwargs)
    return closure


class PumpModel (object):
  bolus_strokes = 20
  basal_strokes = 40
  MMOL_DEFAULT = False
  larger = False
  Ian50Body = 30
  def __init__(self, model, session):
    self.model = model
    self.session = session

  read_model = Task(commands.ReadPumpModel)
  read_status = Task(commands.ReadPumpStatus)
  read_temp_basal = Task(commands.ReadBasalTemp)
  read_settings = Task(commands.ReadSettings)
  read_reservoir = Task(commands.ReadRemainingInsulin)
  read_carb_ratios = Task(commands.ReadCarbRatios512)
  read_bg_targets = Task(commands.ReadBGTargets)
  read_insulin_sensitivies = Task(commands.ReadInsulinSensitivities)
  read_insulin_sensitivities = Task(commands.ReadInsulinSensitivities)
  read_current_glucose_pages = Task(commands.ReadCurGlucosePageNumber)
  read_current_history_pages = Task(commands.ReadCurPageNumber)
  suspend_pump = Task(commands.PumpSuspend)
  resume_pump = Task(commands.PumpResume)
  read_battery_status = Task(commands.ReadBatteryStatus)


  KEYPAD = dict(ESC=commands.KeypadPush.ESC
    , ACT= commands.KeypadPush.ACT
    , UP= commands.KeypadPush.UP
    , DOWN= commands.KeypadPush.DOWN
    , EASY= commands.KeypadPush.EASY
  )
  def press_key (self, key=None, **kwds):
    press = self.KEYPAD.get(key, None)
    err = "Tried to press {key}, but only support {keys}".format(key=key, keys=', '.join(self.KEYPAD.keys( )))
    assert press, err
    # req = press(**kwds)
    resp = self.session.query(press, **kwds)
    packet = resp.getData( )
    raw = str(packet).encode('hex')
    result = dict(raw=raw, received=len(packet) > 1, key=key)

    return result
  def decode_unabsorbed (self, raw):
    doses = [ ]
    while raw and len(raw) > 2:
      head, tail = raw[:3], raw[3:]
      doses.append(self.decode_unabsorbed_component(*head) )
      raw = tail
    return doses

  def decode_unabsorbed_component (self, amount, age, _curve,strokes=40.0):
    curve = ((_curve & 0b110000) << 4)
    unabsorbed = { 'amount': amount/strokes,
                   'age': age + curve,
                   # 'curve': curve,
                 }
    return unabsorbed

  @PageIterator.handler( )
  class iter_glucose_pages (Cursor):
    Info = commands.ReadCurGlucosePageNumber
    Page = commands.ReadGlucoseHistory
    def range (self, info):
      start = int(info['page'])
      end = start - int(info['glucose'])
      return xrange(start, end, -1)
    def find_records (self, response):
      page = cgm.PagedData.Data(response.data, larger=self.inst.larger)
      return reversed(page.decode( ))

  @PageIterator.handler( )
  class iter_history_pages (Cursor):
    Info = commands.ReadCurPageNumber
    Page = commands.ReadHistoryData
    def range (self, info):
      start = 0
      end = int(info)
      return xrange(start, end)
    def find_records (self, response):
      decoder = history.HistoryPage(response.data, self.inst)
      records = decoder.decode( )
      return records

  filter_glucose_date = Task(commands.FilterGlucoseHistory.ISO)
  filter_isig_date = Task(commands.FilterISIGHistory.ISO)

  @Task.handler(commands.ReadHistoryData)
  def read_history_data (self, response):
    decoder = history.HistoryPage(response.data, self)
    records = decoder.decode( )
    return records

  @Task.handler(commands.ReadGlucoseHistory)
  def read_glucose_data (self, response):
    records = [ ]
    page = cgm.PagedData.Data(response.data, larger=self.larger)
    records.extend(page.decode( ))
    return records

  @Task.handler(commands.ReadSettings)
  def my_read_settings (self, response):
    self.settings = response.getData( )
    return self.settings

  @Task.handler(commands.ReadRTC)
  def read_clock (self, response):
    clock = lib.parse.date(response.getData( ))
    return clock

  @Task.handler(commands.SetRTC)
  def _set_clock (self, response):
    raw = response.getData( )
    # TODO: fix mmeowlink's zero responses
    # print "SET RESPONSE RAW", raw, response
    # clock = lib.parse.date(raw)
    return raw
  def set_clock (self, clock=None, **kwds):
    params = commands.SetRTC.fmt_datetime(clock)
    old = self.read_clock( )
    program = dict(requested=dict(clock=clock, params=list(params), old=old))
    results = self._set_clock(params=params, **kwds)
    program.update(raw=list(results))
    program.update(clock=self.read_clock( ))
    return program

  _set_temp_basal = Task(commands.TempBasal.Program)

  _bolus = Task(commands.Bolus)
  strokes_per_unit = 10
  def bolus (self, units=None, **kwds):
    params = self.fmt_bolus_params(units)
    program = dict(requested=dict(units=units, params=list(params)))
    results = self._bolus(params=params, **kwds)
    program.update(**results)
    program.update(**self.read_status( ))
    return program
  def fmt_bolus_params (self, units):
    strokes = int(float(units) * self.strokes_per_unit)
    if (self.larger or self.strokes_per_unit > 10):
      return [lib.HighByte(strokes), lib.LowByte(strokes)]
    return [strokes]

  def set_temp_basal (self, rate=None, duration=None, temp=None, **kwds):
    basals = dict(rate=rate, duration=duration, temp=temp)
    result = self._set_temp_basal(**basals)
    if not result.get('recieved'):
      result.update(requested=basals)
    result.update(**self.read_temp_basal( ))
    return result


class Model508 (PumpModel):
  MMOL_DEFAULT = False
  old6cBody = 31
  # XXX: hack to return something.
  def read_status (self, **kwds):
    number = self.read_model( )
    status = dict(model=number, error="not supported")
    return status

class Model511 (Model508):
  read_basal_profile_std = Task(commands.ReadProfiles511_STD)
  read_basal_profile_a = Task(commands.ReadProfiles511_A)
  read_basal_profile_b = Task(commands.ReadProfiles511_B)

  def read_selected_basal_profile (self, **kwds):
    settings = self.read_settings( )
    selected = settings['selected_pattern']
    patterns = {
        0 : self.read_basal_profile_std
      , 1 : self.read_basal_profile_a
      , 2 : self.read_basal_profile_b
      }
    return patterns[selected](**kwds)


class Model512 (Model511):
  MMOL_DEFAULT = False
  read_basal_profile_std = Task(commands.ReadProfile_STD512)
  read_basal_profile_a = Task(commands.ReadProfile_A512)
  read_basal_profile_b = Task(commands.ReadProfile_B512)


class Model515 (Model512):
  MMOL_DEFAULT = False
  read_bg_targets = Task(commands.ReadBGTargets515)
  read_status = Task(commands.ReadPumpStatus)
  pass

class Model715 (Model515):
  MMOL_DEFAULT = False
  pass

class Model522 (Model515):
  MMOL_DEFAULT = False
  old6cBody = 38
  pass

class Model722 (Model522):
  MMOL_DEFAULT = False
  pass

class Model523 (Model522):
  strokes_per_unit = 40
  larger = True
  read_carb_ratios = Task(commands.ReadCarbRatios)
  read_reservoir = Task(commands.ReadRemainingInsulin523)
  read_settings = Task(commands.ReadSettings523)

class Model723 (Model523):
  pass

class Model530 (Model523):
  Ian50Body = 34
  pass

class Model730 (Model530):
  Ian50Body = 34
  pass

class Model540 (Model530):
  MMOL_DEFAULT = True
  pass

class Model740 (Model540):
  MMOL_DEFAULT = True
  pass

class Model551 (Model540):
  MMOL_DEFAULT = True
  pass

class Model751 (Model551):
  MMOL_DEFAULT = True
  pass

class Model554 (Model551):
  MMOL_DEFAULT = True
  pass

class Model754 (Model554):
  MMOL_DEFAULT = True
  pass

known = {
  '508': Model508
, '511': Model511
, '512': Model512
, '515': Model515
, '522': Model522
, '523': Model523
, '530': Model530
, '540': Model540
, '551': Model551
, '554': Model554
, '715': Model715
, '722': Model722
, '723': Model723
, '723': Model723
, '730': Model730
, '740': Model740
, '751': Model751
, '754': Model754
}

def lookup (model, session):
  klass = known.get(model, PumpModel)
  return klass(model, session)

